"""
Reading Offline events
======================

The following example shows how to access events data in an offline ROOT file, which is
written by aanet software.

Note: the offline file used here has MC offline data and was intentionaly reduced
to 10 events.
"""
import km3io as ki
from km3net_testdata import data_path

#####################################################
# First, pass a filename to the `OfflineReader` class to open the file.
# Note that only some meta information is read into memory.

r = ki.OfflineReader(data_path("offline/numucc.root"))


#####################################################
# Accessing the file header
# -------------------------
# Note that not all file headers are supported, so don't be surprised if
# nothing is returned when the file header is called (this can happen if your file
# was produced with old versions of aanet).

h = r.header
print(h)


#####################################################
# Accessing the events data
# -------------------------
# Note that not all data is loaded in memory (again), so printing
# events will only return how many events were found in the file.

print(r.events)


#####################################################
# to explore the events keys:

keys = r.events.keys()
print(keys)

#####################################################
# to access the number of hits associated with each event:

n_hits = r.events.n_hits
print(n_hits)

#####################################################
# to access the number of tracks associated with each event:

n_tracks = r.events.n_tracks
print(n_tracks)

#####################################################
# to access the number of mc hits associated with each event:

n_mc_hits = r.events.n_mc_hits
print(n_mc_hits)

#####################################################
# to access the number of mc tracks associated with each event:

n_mc_tracks = r.events.n_mc_tracks
print(n_mc_tracks)

#####################################################
# to access the overlays:

overlays = r.events.overlays
print(overlays)

#####################################################
# That's it! you can access any key of your interest in the events
# keys in the exact same way.


#####################################################
# item selection in events data
# -----------------------------
# events can be selected as you would select an item from a numpy array.
# for example, to select the mc_hits in event 0:

print(r.events[0].mc_hits)

#####################################################
# or:

print(r.events.mc_hits[0])


#####################################################
# slicing of events
# -----------------
# you can select a slice of events data. For example, to select the number
# of mc hits in the first 5 events:

print(r.events.n_mc_hits[0:5])

#####################################################
# or:

print(r.events[0:5].n_mc_hits)


#####################################################
# you can apply masks to events data as you would do with numpy arrays.
# For example, to select the number of hits higher than 50:

mask = r.events.n_mc_hits > 50

print(r.events.n_mc_hits[mask])

#####################################################
# or:

print(r.events.n_mc_tracks[mask])



import os
import re
import unittest
import inspect
import awkward as ak

from km3net_testdata import data_path

from km3io.gseagen import GSGReader

GSG_READER = GSGReader(data_path("gseagen/gseagen.root"))

AWKWARD_STR_CLASSES = [
    s[1] for s in inspect.getmembers(ak.behaviors.string, inspect.isclass)
]


class TestGSGHeader(unittest.TestCase):
    def setUp(self):
        self.header = GSG_READER.header

    def test_str_byte_type(self):
        assert type(self.header["gSeaGenVer"]) in AWKWARD_STR_CLASSES
        assert type(self.header["GenieVer"]) in AWKWARD_STR_CLASSES
        assert type(self.header["gSeaGenVer"]) in AWKWARD_STR_CLASSES
        assert type(self.header["InpXSecFile"]) in AWKWARD_STR_CLASSES
        assert type(self.header["Flux1"]) in AWKWARD_STR_CLASSES
        assert type(self.header["Flux2"]) in AWKWARD_STR_CLASSES

    def test_values(self):
        assert self.header["RunNu"] == 1
        assert self.header["RanSeed"] == 3662074
        self.assertAlmostEqual(self.header["NTot"], 1000.0)
        self.assertAlmostEqual(self.header["EvMin"], 5.0)
        self.assertAlmostEqual(self.header["EvMax"], 50.0)
        self.assertAlmostEqual(self.header["CtMin"], -1.0)
        self.assertAlmostEqual(self.header["CtMax"], 1.0)
        self.assertAlmostEqual(self.header["Alpha"], 1.4)
        assert self.header["NBin"] == 1
        self.assertAlmostEqual(self.header["Can1"], 0.0)
        self.assertAlmostEqual(self.header["Can2"], 476.5)
        self.assertAlmostEqual(self.header["Can3"], 403.4)
        self.assertAlmostEqual(self.header["OriginX"], 0.0)
        self.assertAlmostEqual(self.header["OriginY"], 0.0)
        self.assertAlmostEqual(self.header["OriginZ"], 0.0)
        self.assertAlmostEqual(self.header["HRock"], 93.03036681)
        self.assertAlmostEqual(self.header["HSeaWater"], 684.39143353)
        self.assertAlmostEqual(self.header["RVol"], 611.29143353)
        self.assertAlmostEqual(self.header["HBedRock"], 0.0)
        self.assertAlmostEqual(self.header["NAbsLength"], 3.0)
        self.assertAlmostEqual(self.header["AbsLength"], 93.33)
        self.assertAlmostEqual(self.header["SiteDepth"], 2425.0)
        self.assertAlmostEqual(self.header["SiteLatitude"], 0.747)
        self.assertAlmostEqual(self.header["SiteLongitude"], 0.10763)
        self.assertAlmostEqual(self.header["SeaBottomRadius"], 6368000.0)
        assert round(self.header["GlobalGenWeight"] - 6.26910765e08, 0) == 0
        self.assertAlmostEqual(self.header["RhoSW"], 1.03975)
        self.assertAlmostEqual(self.header["RhoSR"], 2.65)
        self.assertAlmostEqual(self.header["TGen"], 31556926.0)
        assert not self.header["PropMode"]
        assert self.header["NNu"] == 2
        self.assertListEqual(self.header["NuList"].tolist(), [-14, 14])


class TestGSGEvents(unittest.TestCase):
    def setUp(self):
        self.events = GSG_READER.events

    def test_event_single_values(self):
        event = self.events[0]
        assert 1 == event.iEvt
        self.assertAlmostEqual(event.PScale, 1.0)
        assert event.TargetA == 28
        assert event.TargetZ == 14
        assert event.InterId == 2
        assert event.ScattId == 3
        self.assertAlmostEqual(event.Bx, 0.21398980096236023)
        self.assertAlmostEqual(event.By, 0.7736389792036026)
        assert event.VerInCan == 0
        self.assertAlmostEqual(event.WaterXSec, 9.814545541159586e-42)
        self.assertAlmostEqual(event.WaterIntLen, 169191612204.44382)
        self.assertAlmostEqual(event.PEarth, 0.9998206835107583)
        self.assertAlmostEqual(event.ColumnDepth, 31352421806833.758)
        self.assertAlmostEqual(event.XSecMean, 9.49810757204515e-42)
        self.assertAlmostEqual(event.Agen, 934842.6115446005)
        self.assertAlmostEqual(event.WGen, 480360599.4207591)
        self.assertAlmostEqual(event.WEvt, 5.629637364719835)
        self.assertAlmostEqual(event.E_nu, 29.210801854810974)
        assert event.Pdg_nu == -14
        self.assertAlmostEqual(event.Vx_nu, 202.86806811399845)
        self.assertAlmostEqual(event.Vy_nu, 226.64032038584787)
        self.assertAlmostEqual(event.Vz_nu, -0.5526929982180058)
        self.assertAlmostEqual(event.Dx_nu, -0.7554430076659393)
        self.assertAlmostEqual(event.Dy_nu, 0.3034298552575011)
        self.assertAlmostEqual(event.Dz_nu, 0.5807204018346965)
        self.assertAlmostEqual(event.T_nu, 0.0)
        self.assertAlmostEqual(event.E_pl, 6.648776215872709)
        assert event.Pdg_pl == -13
        self.assertAlmostEqual(event.Vx_pl, 202.86806811399845)
        self.assertAlmostEqual(event.Vy_pl, 226.64032038584787)
        self.assertAlmostEqual(event.Vz_pl, -0.5526929982180058)
        self.assertAlmostEqual(event.Dx_pl, -0.8781745795144783)
        self.assertAlmostEqual(event.Dy_pl, 0.2818890466120743)
        self.assertAlmostEqual(event.Dz_pl, 0.3864556550171106)
        self.assertAlmostEqual(event.T_pl, 0.0)
        assert event.NTracks == 1

    def test_event_list_values(self):
        event = self.events[1]
        self.assertListEqual(event.Id_tr.tolist(), [4, 5, 10, 11, 12])
        self.assertListEqual(event.Pdg_tr.tolist(), [22, -13, 2112, -211, 111])
        [
            self.assertAlmostEqual(x, y)
            for x, y in zip(
                event.E_tr, [0.00618, 4.88912206, 2.33667201, 1.0022909, 1.17186997]
            )
        ]
        [
            self.assertAlmostEqual(x, y)
            for x, y in zip(
                event.Vx_tr,
                [
                    -337.67895799,
                    -337.67895799,
                    -337.67895799,
                    -337.67895799,
                    -337.67895799,
                ],
            )
        ]
        [
            self.assertAlmostEqual(x, y)
            for x, y in zip(
                event.Vy_tr,
                [
                    -203.90999969,
                    -203.90999969,
                    -203.90999969,
                    -203.90999969,
                    -203.90999969,
                ],
            )
        ]
        [
            self.assertAlmostEqual(x, y)
            for x, y in zip(
                event.Vz_tr,
                [416.08845294, 416.08845294, 416.08845294, 416.08845294, 416.08845294],
            )
        ]
        [
            self.assertAlmostEqual(x, y)
            for x, y in zip(
                event.Dx_tr,
                [0.06766196, -0.63563065, -0.70627586, -0.76364544, -0.80562216],
            )
        ]
        [
            self.assertAlmostEqual(x, y)
            for x, y in zip(
                event.Dy_tr,
                [0.33938809, -0.4846643, 0.50569058, -0.04136113, 0.10913917],
            )
        ]
        [
            self.assertAlmostEqual(x, y)
            for x, y in zip(
                event.Dz_tr,
                [-0.93820978, -0.6008945, -0.49543056, -0.64430963, -0.58228994],
            )
        ]
        [self.assertAlmostEqual(x, y) for x, y in zip(event.T_tr, 5 * [0.0])]



#!/usr/bin/env python3
from collections import namedtuple
import numpy as np
import awkward as ak
import uproot

from .tools import unfold_indices

import logging

log = logging.getLogger("km3io.rootio")


class EventReader:
    """reader for offline ROOT files"""

    event_path = None
    item_name = "Event"
    skip_keys = []  # ignore these subbranches, even if they exist
    aliases = {}  # top level aliases -> {fromkey: tokey}
    nested_branches = {}
    nested_aliases = {}

    def __init__(
        self,
        f,
        index_chain=None,
        step_size=2000,
        keys=None,
        aliases=None,
        nested_branches=None,
        event_ctor=None,
    ):
        """EventReader base class

        Parameters
        ----------
        f : str or uproot4.reading.ReadOnlyDirectory (from uproot4.open)
            Path to the file of interest or uproot4 filedescriptor.
        step_size : int, optional
            Number of events to read into the cache when iterating.
            Choosing higher numbers may improve the speed but also increases
            the memory overhead.
        index_chain : list, optional
            Keeps track of index chaining.
        keys : list or set, optional
            Branch keys.
        aliases : dict, optional
            Branch key aliases.
        event_ctor : class or namedtuple, optional
            Event constructor.

        """
        if isinstance(f, str):
            self._fobj = uproot.open(f)
            self._filepath = f
        elif isinstance(f, uproot.reading.ReadOnlyDirectory):
            self._fobj = f
            self._filepath = f._file.file_path
        else:
            raise TypeError("Unsupported file descriptor.")
        self._step_size = step_size
        self._uuid = self._fobj._file.uuid
        self._iterator_index = 0
        self._keys = keys
        self._event_ctor = event_ctor
        self._index_chain = [] if index_chain is None else index_chain

        if aliases is not None:
            self.aliases = aliases
        if nested_branches is not None:
            self.nested_branches = nested_branches

        if self._keys is None:
            self._initialise_keys()

        if self._event_ctor is None:
            self._event_ctor = namedtuple(
                self.item_name,
                set(
                    list(self.keys())
                    + list(self.aliases)
                    + list(self.nested_branches)
                    + list(self.nested_aliases)
                ),
            )

    def _initialise_keys(self):
        skip_keys = set(self.skip_keys)
        all_keys = set(self._fobj[self.event_path].keys())
        toplevel_keys = set(k.split("/")[0] for k in all_keys)
        valid_aliases = {}
        for fromkey, tokey in self.aliases.items():
            if tokey in all_keys:
                valid_aliases[fromkey] = tokey
        self.aliases = valid_aliases
        keys = (toplevel_keys - skip_keys).union(
            list(valid_aliases) + list(self.nested_aliases)
        )
        for key in list(self.nested_branches) + list(self.nested_aliases):
            keys.add("n_" + key)
        # self._grouped_branches = {k for k in toplevel_keys - skip_keys if isinstance(self._fobj[self.event_path][k].interpretation, uproot.AsGrouped)}
        valid_nested_branches = {}
        for nested_key, aliases in self.nested_branches.items():
            if nested_key in toplevel_keys:
                valid_nested_branches[nested_key] = {}
                subbranch_keys = self._fobj[self.event_path][nested_key].keys()
                for fromkey, tokey in aliases.items():
                    if tokey in subbranch_keys:
                        valid_nested_branches[nested_key][fromkey] = tokey
        self.nested_branches = valid_nested_branches
        self._keys = keys

    def __dir__(self):
        """Tab completion in IPython"""
        return list(self.keys()) + ["header"]

    def keys(self):
        """Returns all accessible branch keys, without the skipped ones."""
        return self._keys

    @property
    def events(self):
        # TODO: deprecate this, since `self` is already the container type
        return iter(self)

    def _keyfor(self, key):
        """Return the correct key for a given alias/key"""
        return self.nested_aliases.get(key, key)

    def __getattr__(self, attr):
        attr = self._keyfor(attr)
        # if attr in self.keys() or (attr.startswith("n_") and self._keyfor(attr.split("n_")[1]) in self._grouped_branches):
        if attr in self.keys():
            return self.__getitem__(attr)
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{attr}'"
        )

    def __getitem__(self, key):
        # indexing
        # TODO: maybe just propagate everything to awkward and let it deal
        # with the type?
        if isinstance(
            key, (slice, int, np.int32, np.int64, list, np.ndarray, ak.Array)
        ):
            if isinstance(key, (int, np.int32, np.int64)):
                key = int(key)
            return self.__class__(
                self._fobj,
                index_chain=self._index_chain + [key],
                step_size=self._step_size,
                aliases=self.aliases,
                nested_branches=self.nested_branches,
                keys=self.keys(),
                event_ctor=self._event_ctor,
            )
        # group counts, for e.g. n_events, n_hits etc.
        if isinstance(key, str) and key.startswith("n_"):
            key = self._keyfor(key.split("n_")[1])
            arr = self._fobj[self.event_path][key].array(uproot.AsDtype(">i4"))
            return unfold_indices(arr, self._index_chain)

        key = self._keyfor(key)
        branch = self._fobj[self.event_path]
        # These are special branches which are nested, like hits/trks/mc_trks
        # We are explicitly grabbing just a predefined set of subbranches
        # and also alias them to be backwards compatible (and attribute-accessible)
        if key in self.nested_branches:
            fields = []
            # some fields are not always available, like `usr_names`
            for to_field, from_field in self.nested_branches[key].items():
                if from_field in branch[key].keys():
                    fields.append(to_field)
            log.debug(fields)
            return Branch(
                branch[key], fields, self.nested_branches[key], self._index_chain
            )
        else:
            return unfold_indices(
                branch[self.aliases.get(key, key)].array(), self._index_chain
            )

    def __iter__(self, chunkwise=False):
        self._events = self._event_generator(chunkwise=chunkwise)
        return self

    def _get_iterator_limits(self):
        """Determines start and stop, used for event iteration"""
        if len(self._index_chain) > 1:
            raise NotImplementedError(
                "iteration is currently not supported with nested slices"
            )
        if self._index_chain:
            s = self._index_chain[0]
            if not isinstance(s, slice):
                raise NotImplementedError("iteration is only supported with slices")
            if s.step is None or s.step == 1:
                start = s.start
                stop = s.stop
            else:
                raise NotImplementedError(
                    "iteration is only supported with single steps"
                )
        else:
            start = None
            stop = None
        return start, stop

    def _event_generator(self, chunkwise=False):
        start, stop = self._get_iterator_limits()

        if chunkwise:
            raise NotImplementedError("iterating over chunks is not implemented yet")

        events = self._fobj[self.event_path]
        group_count_keys = set(
            k for k in self.keys() if k.startswith("n_")
        )  # extra keys to make it easy to count subbranch lengths
        log.debug("group_count_keys: %s", group_count_keys)
        keys = set(
            list(
                set(self.keys())
                - set(self.nested_branches.keys())
                - set(self.nested_aliases)
                - group_count_keys
            )
            + list(self.aliases.keys())
        )  # all top-level keys for regular branches
        log.debug("keys: %s", keys)
        log.debug("aliases: %s", self.aliases)
        events_it = events.iterate(
            keys,
            aliases=self.aliases,
            step_size=self._step_size,
            entry_start=start,
            entry_stop=stop,
        )
        nested = []
        nested_keys = (
            self.nested_branches.keys()
        )  # dict-key ordering is an implementation detail
        log.debug("nested_keys: %s", nested_keys)
        for key in nested_keys:
            nested.append(
                events[key].iterate(
                    self.nested_branches[key].keys(),
                    aliases=self.nested_branches[key],
                    step_size=self._step_size,
                    entry_start=start,
                    entry_stop=stop,
                )
            )
        group_counts = {}
        for key in group_count_keys:
            group_counts[key] = iter(self[key])

        log.debug("group_counts: %s", group_counts)
        for event_set, *nested_sets in zip(events_it, *nested):
            for _event, *nested_items in zip(event_set, *nested_sets):
                data = {}
                for k in keys:
                    data[k] = _event[k]
                for (k, i) in zip(nested_keys, nested_items):
                    data[k] = i
                for tokey, fromkey in self.nested_aliases.items():
                    data[tokey] = data[fromkey]
                for key in group_counts:
                    data[key] = next(group_counts[key])
                yield self._event_ctor(**data)

    def __next__(self):
        return next(self._events)

    def __len__(self):
        if not self._index_chain:
            return self._fobj[self.event_path].num_entries
        elif isinstance(self._index_chain[-1], (int, np.int32, np.int64)):
            if len(self._index_chain) == 1:
                # TODO: not sure why this is needed at all, it's too late...
                return 1
                # try:
                #     return len(self[:])
                # except IndexError:
                #     return 1
            return 1
        else:
            # ignore the usual index magic and access `id` directly
            return len(
                unfold_indices(
                    self._fobj[self.event_path]["id"].array(), self._index_chain
                )
            )

    def __actual_len__(self):
        """The raw number of events without any indexing/slicing magic"""
        return len(self._fobj[self.event_path]["id"].array())

    def __repr__(self):
        length = len(self)
        actual_length = self.__actual_len__()
        return f"{self.__class__.__name__} ({length}{'/' + str(actual_length) if length < actual_length else ''} events)"

    @property
    def uuid(self):
        return self._uuid

    def close(self):
        self._fobj.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class Branch:
    """Helper class for nested branches likes tracks/hits"""

    def __init__(self, branch, fields, aliases, index_chain):
        self._branch = branch
        self.fields = fields
        self._aliases = aliases
        self._index_chain = index_chain

    def __dir__(self):
        """Tab completion in IPython"""
        return list(self.fields)

    def __getattr__(self, attr):
        if attr not in self._aliases:
            raise AttributeError(
                f"No field named {attr}. Available fields: {self.fields}"
            )
        key = self._aliases[attr]

        if self._index_chain:
            idx0 = self._index_chain[0]
            if isinstance(idx0, (int, np.int32, np.int64)):
                # optimise single-element and slice lookups
                start = idx0
                stop = idx0 + 1
                arr = ak.flatten(
                    self._branch[key].array(entry_start=start, entry_stop=stop)
                )
                return unfold_indices(arr, self._index_chain[1:])
            if isinstance(idx0, slice):
                if idx0.step is None or idx0.step == 1:
                    start = idx0.start
                    stop = idx0.stop
                    arr = self._branch[key].array(entry_start=start, entry_stop=stop)
                    return unfold_indices(arr, self._index_chain[1:])

        return unfold_indices(self._branch[key].array(), self._index_chain)

    def __getitem__(self, key):
        return self.__class__(
            self._branch, self.fields, self._aliases, self._index_chain + [key]
        )

    def __len__(self):
        if not self._index_chain:
            return self._branch.num_entries
        elif isinstance(self._index_chain[-1], (int, np.int32, np.int64)):
            if len(self._index_chain) == 1:
                return 1
                # try:
                #     return len(self[:])
                # except IndexError:
                #     return 1
            return 1
        else:
            # ignore the usual index magic and access `id` directly
            return len(self.id)

    def __actual_len__(self):
        """The raw number of events without any indexing/slicing magic"""
        return len(self._branch[self._aliases["id"]].array())

    def __repr__(self):
        length = len(self)
        actual_length = self.__actual_len__()
        return f"{self.__class__.__name__} ({length}{'/' + str(actual_length) if length < actual_length else ''} {self._branch.name})"

    @property
    def ndim(self):
        if not self._index_chain:
            return 2
        elif any(isinstance(i, (int, np.int32, np.int64)) for i in self._index_chain):
            return 1
        return 2



"""
Reading Online Data
===================

The following example shows how to access hits in a ROOT file which is coming
from the detector and written by the `JDataWriter` application.

Such a file is usually called "KM3NET_00000001_00000002.root", where the first
number is the detector ID and the second the run number.
"""
import km3io as ki
from km3net_testdata import data_path

#####################################################
# Accessing the event tree
# ------------------------
# Just pass a filename to the reader class and get access to the event tree
# with:


f = ki.OnlineReader(data_path("online/km3net_online.root"))

#####################################################
# Note that only some meta information is read into memory.
#
# Printing it will simply tell you how many events it has found. Again, nothing
# else is read yet:

print(f.events)

#####################################################
# Now let's look at the hits data:

print(f.events[0].snapshot_hits.tot)

#####################################################
# the resulting arrays are numpy arrays.

#####################################################
# Reading SummarySlices
# ---------------------
# The following example shows how to access summary slices, in particular the DOM
# IDs of the slice with the index 0:

dom_ids = f.summaryslices.slices[0].dom_id

print(dom_ids)

#####################################################
# The .dtype attribute (or in general, <TAB> completion) is useful to find out
# more about the field structure:

print(f.summaryslices.headers.dtype)

#####################################################
# To read the frame index:

print(f.summaryslices.headers.frame_index)

#####################################################
# The resulting array is a ChunkedArray which is an extended version of a
# numpy array and behaves like one.

#####################################################
# Reading TimeSlices
# ------------------
# To be continued.
#



"""
Reading Offline tracks
======================

The following example shows how to access tracks data in an offline ROOT file.

Note: the offline files used here were intentionaly reduced to 10 events.
"""
import km3io as ki
from km3net_testdata import data_path

#####################################################
# We open the file using the

f = ki.OfflineReader(data_path("offline/numucc.root"))

#####################################################
# To access offline tracks/mc_tracks data:

f.tracks
f.mc_tracks

#####################################################
# Note that no data is loaded in memory at this point, so printing
# tracks will only return how many sub-branches (corresponding to
# events) were found.

f.tracks

#####################################################
# same for mc hits

f.mc_tracks

#####################################################
# Accessing the tracks/mc_tracks keys
# -----------------------------------
# to explore the reconstructed tracks fields:

f.tracks.fields

#####################################################
# the same for MC tracks

f.mc_tracks.fields

#####################################################
# Accessing tracks data
# ---------------------
# each field will return a nested `awkward.Array` and load everything into
# memory, so be careful if you are working with larger files.

f.tracks.E

######################################################
# The z direction of all reconstructed tracks

f.tracks.dir_z

######################################################
# The likelihoods

f.tracks.lik


#####################################################
# To select just a single event or a subset of events, use the indices or slices.
# The following will access all tracks and their fields
# of the third event (0 is the first):

f[2].tracks

######################################################
# The z direction of all tracks in the third event:

f[2].tracks.dir_z


#####################################################
# while here, we select the first 3 events. Notice that all fields will return
# nested arrays, as we have seem above where all events were selected.

f[:3]

######################################################
# All tracks for the first three events

f[:3].tracks

######################################################
# The z directions of all tracks of the first three events

f[:3].tracks.dir_z

#####################################################
# or events from 3 and 5 (again, 0 indexing):


f[2:5]

######################################################
# the tracks of those events

f[2:5].tracks

######################################################
# and just the z directions of those

f[2:5].tracks.dir_z



#!/usr/bin/env python3

from km3io._definitions.mc_header import data as mc_header

from km3io._definitions.trigger import data as trigger
from km3io._definitions.fitparameters import data as fitparameters
from km3io._definitions.reconstruction import data as reconstruction
from km3io._definitions.w2list_genhen import data as w2list_genhen
from km3io._definitions.w2list_gseagen import data as w2list_gseagen


class AttrDict(dict):
    """A dictionary which allows access to its key through attributes."""

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


trigger = AttrDict(trigger)
fitparameters = AttrDict(fitparameters)
reconstruction = AttrDict(reconstruction)
w2list_genhen = AttrDict(w2list_genhen)
w2list_gseagen = AttrDict(w2list_gseagen)

trigger_idx = {v: k for k, v in trigger.items()}
fitparameters_idx = {v: k for k, v in fitparameters.items()}
reconstruction_idx = {v: k for k, v in reconstruction.items()}
w2list_genhen_idx = {v: k for k, v in w2list_genhen.items()}
w2list_gseagen_idx = {v: k for k, v in w2list_gseagen.items()}



#!/usr/bin/env python
# coding=utf-8
# Filename: gseagen.py
# Author: Johannes Schumann <jschumann@km3net.de>

import warnings
from .rootio import EventReader
from .tools import cached_property


class GSGReader(EventReader):
    """reader for gSeaGen ROOT files"""

    header_key = "Header"
    event_path = "Events"
    skip_keys = [header_key]

    @cached_property
    def header(self):
        if self.header_key in self._fobj:
            header = {}
            for k, v in self._fobj[self.header_key].items():
                v = v.array()[0]
                if isinstance(v, bytes):
                    try:
                        v = v.decode("utf-8")
                    except UnicodeDecodeError:
                        pass
                header[k] = v
            return header
        else:
            warnings.warn("Your file header has an unsupported format")



from pkg_resources import get_distribution, DistributionNotFound

version = get_distribution(__name__).version

from .offline import OfflineReader
from .online import OnlineReader
from .gseagen import GSGReader



#!/usr/bin/env python3
data = {
    "DAQ": "livetime",
    "seed": "program level iseed",
    "PM1_type_area": "type area TTS",
    "PDF": "i1 i2",
    "model": "interaction muon scattering numberOfEnergyBins",
    "can": "zmin zmax r",
    "genvol": "zmin zmax r volume numberOfEvents",
    "merge": "time gain",
    "coord_origin": "x y z",
    "translate": "x y z",
    "genhencut": "gDir Emin",
    "k40": "rate time",
    "norma": "primaryFlux numberOfPrimaries",
    "livetime": "numberOfSeconds errorOfSeconds",
    "flux": "type key file_1 file_2",
    "spectrum": "alpha",
    "fixedcan": "xcenter ycenter zmin zmax radius",
    "start_run": "run_id",
}

for key in "cut_primary cut_seamuon cut_in cut_nu".split():
    data[key] = "Emin Emax cosTmin cosTmax"

for key in "generator physics simul".split():
    data[key] = "program version date time"

for key in data.keys():
    data[key] = data[key].split()



# -*- coding: utf-8 -*-
"""
KM3NeT Data Definitions v1.2.4
https://git.km3net.de/common/km3net-dataformat
"""

# reconstruction
data = {
    "JPP_RECONSTRUCTION_TYPE": 4000,
    "JMUONFIT": 0,
    "JMUONBEGIN": 0,
    "JMUONPREFIT": 1,
    "JMUONSIMPLEX": 2,
    "JMUONGANDALF": 3,
    "JMUONENERGY": 4,
    "JMUONSTART": 5,
    "JLINEFIT": 6,
    "JMUONEND": 99,
    "JSHOWERFIT": 100,
    "JSHOWERBEGIN": 100,
    "JSHOWERPREFIT": 101,
    "JSHOWERPOSITIONFIT": 102,
    "JSHOWERCOMPLETEFIT": 103,
    "JSHOWER_BJORKEN_Y": 104,
    "JSHOWERENERGYPREFIT": 105,
    "JSHOWERPOINTSIMPLEX": 106,
    "JSHOWERDIRECTIONPREFIT": 107,
    "JSHOWEREND": 199,
    "DUSJ_RECONSTRUCTION_TYPE": 200,
    "DUSJSHOWERBEGIN": 200,
    "DUSJSHOWERPREFIT": 201,
    "DUSJSHOWERPOSITIONFIT": 202,
    "DUSJSHOWERCOMPLETEFIT": 203,
    "DUSJSHOWEREND": 299,
    "AANET_RECONSTRUCTION_TYPE": 101,
    "AASHOWERBEGIN": 300,
    "AASHOWERFITPREFIT": 302,
    "AASHOWERFITPOSITIONFIT": 303,
    "AASHOWERFITDIRECTIONENERGYFIT": 304,
    "AASHOWEREND": 399,
    "JUSERBEGIN": 1000,
    "JMUONVETO": 1001,
    "JMUONPATH": 1003,
    "JMCEVT": 1004,
    "JUSEREND": 1099,
    "RECTYPE_UNKNOWN": -1,
    "RECSTAGE_UNKNOWN": -1,
}



# -*- coding: utf-8 -*-
"""
KM3NeT Data Definitions v1.2.4
https://git.km3net.de/common/km3net-dataformat
"""

# w2list_genhen
data = {
    "W2LIST_GENHEN_GLOBAL_GEN_WEIGHT": 0,
    "W2LIST_GENHEN_EG": 1,
    "W2LIST_GENHEN_SIG": 2,
    "W2LIST_GENHEN_COLUMN_DEPTH": 3,
    "W2LIST_GENHEN_P_EARTH": 4,
    "W2LIST_GENHEN_REFF": 5,
    "W2LIST_GENHEN_BX": 7,
    "W2LIST_GENHEN_BY": 8,
    "W2LIST_GENHEN_ICHAN": 9,
    "W2LIST_GENHEN_CC": 10,
}



from collections import namedtuple
import itertools
import os
import re
import unittest

from km3net_testdata import data_path

from km3io.online import (
    OnlineReader,
    get_rate,
    has_udp_trailer,
    get_udp_max_sequence_number,
    get_channel_flags,
    get_number_udp_packets,
)
from km3io.tools import to_num

ONLINE_FILE = data_path("online/km3net_online.root")


class TestOnlineReaderContextManager(unittest.TestCase):
    def test_context_manager(self):
        with OnlineReader(ONLINE_FILE) as r:
            assert r._filename == ONLINE_FILE


class TestUUID(unittest.TestCase):
    def test_uuid(self):
        assert OnlineReader(ONLINE_FILE).uuid == "00010c85603008c611ea971772f09e86beef"


class TestOnlineEvents(unittest.TestCase):
    def setUp(self):
        self.events = OnlineReader(ONLINE_FILE).events

    def test_index_lookup(self):
        assert 3 == len(self.events)

    def test_str(self):
        assert re.match(".*events.*3", str(self.events))

    def test_repr(self):
        assert re.match(".*events.*3", self.events.__repr__())


class TestOnlineEvent(unittest.TestCase):
    def setUp(self):
        self.event = OnlineReader(ONLINE_FILE).events[0]

    def test_str(self):
        assert re.match(".*event.*96.*snapshot.*18.*triggered", str(self.event))

    def test_repr(self):
        assert re.match(".*event.*96.*snapshot.*18.*triggered", self.event.__repr__())


class TestOnlineEventsSnapshotHits(unittest.TestCase):
    def setUp(self):
        self.events = OnlineReader(ONLINE_FILE).events
        self.lengths = {0: 96, 1: 124, -1: 78}
        self.total_item_count = 298

    def test_reading_snapshot_hits(self):
        hits = self.events.snapshot_hits

        for event_id, length in self.lengths.items():
            assert length == len(hits[event_id].dom_id)
            assert length == len(hits[event_id].channel_id)
            assert length == len(hits[event_id].time)

    def test_total_item_counts(self):
        hits = self.events.snapshot_hits

        assert self.total_item_count == sum(hits.dom_id.count())
        assert self.total_item_count == sum(hits.channel_id.count())
        assert self.total_item_count == sum(hits.time.count())

    def test_data_values(self):
        hits = self.events.snapshot_hits

        self.assertListEqual(
            [806451572, 806451572, 806455814], list(hits.dom_id[0][:3])
        )
        self.assertListEqual([10, 13, 0], list(hits.channel_id[0][:3]))
        self.assertListEqual([30733918, 30733916, 30733256], list(hits.time[0][:3]))

    def test_channel_ids_have_valid_values(self):
        hits = self.events.snapshot_hits

        # channel IDs are always between [0, 30]
        assert all(c >= 0 for c in hits.channel_id.min())
        assert all(c < 31 for c in hits.channel_id.max())


class TestOnlineEventsTriggeredHits(unittest.TestCase):
    def setUp(self):
        self.events = OnlineReader(ONLINE_FILE).events
        self.lengths = {0: 18, 1: 53, -1: 9}
        self.total_item_count = 80

    def test_data_lengths(self):
        hits = self.events.triggered_hits

        for event_id, length in self.lengths.items():
            assert length == len(hits[event_id].dom_id)
            assert length == len(hits[event_id].channel_id)
            assert length == len(hits[event_id].time)
            assert length == len(hits[event_id].trigger_mask)

    def test_total_item_counts(self):
        hits = self.events.triggered_hits

        assert self.total_item_count == sum(hits.dom_id.count())
        assert self.total_item_count == sum(hits.channel_id.count())
        assert self.total_item_count == sum(hits.time.count())

    def test_data_values(self):
        hits = self.events.triggered_hits

        self.assertListEqual(
            [806451572, 806451572, 808432835], list(hits.dom_id[0][:3])
        )
        self.assertListEqual([10, 13, 1], list(hits.channel_id[0][:3]))
        self.assertListEqual([30733918, 30733916, 30733429], list(hits.time[0][:3]))
        self.assertListEqual([16, 16, 4], list(hits.trigger_mask[0][:3]))

    def test_channel_ids_have_valid_values(self):
        hits = self.events.triggered_hits

        # channel IDs are always between [0, 30]
        assert all(c >= 0 for c in hits.channel_id.min())
        assert all(c < 31 for c in hits.channel_id.max())


class TestTimeslices(unittest.TestCase):
    def setUp(self):
        self.ts = OnlineReader(ONLINE_FILE).timeslices

    def test_data_lengths(self):
        assert 3 == len(self.ts._timeslices["L1"][0])
        assert 3 == len(self.ts._timeslices["SN"][0])
        with self.assertRaises(KeyError):
            assert 0 == len(self.ts._timeslices["L2"][0])
        with self.assertRaises(KeyError):
            assert 0 == len(self.ts._timeslices["L0"][0])

    def test_streams(self):
        self.ts.stream("L1", 0)
        self.ts.stream("SN", 0)

    def test_reading_frames(self):
        assert 8 == len(self.ts.stream("SN", 1).frames[808447186])

    def test_str(self):
        s = str(self.ts)
        assert "L1" in s
        assert "SN" in s


class TestTimeslice(unittest.TestCase):
    def setUp(self):
        self.ts = OnlineReader(ONLINE_FILE).timeslices
        self.n_frames = {"L1": [69, 69, 69], "SN": [64, 66, 68]}

    def test_str(self):
        for stream, n_frames in self.n_frames.items():
            print(stream, n_frames)
            for i in range(len(n_frames)):
                s = str(self.ts.stream(stream, i))
                assert re.match("{}.*{}".format(stream, n_frames[i]), s)


class TestSummaryslices(unittest.TestCase):
    def setUp(self):
        self.ss = OnlineReader(ONLINE_FILE).summaryslices

    def test_headers(self):
        assert 3 == len(self.ss.headers)
        self.assertListEqual([44, 44, 44], list(self.ss.headers.detector_id))
        self.assertListEqual([6633, 6633, 6633], list(self.ss.headers.run))
        self.assertListEqual([126, 127, 128], list(self.ss.headers.frame_index))
        assert 806451572 == self.ss.slices[0].dom_id[0]

    def test_slices(self):
        assert 3 == len(self.ss.slices)

    def test_rates(self):
        assert 3 == len(self.ss.rates)

    def test_fifo(self):
        s = self.ss.slices[0]
        dct_fifo_stat = {
            808981510: True,
            808981523: False,
            808981672: False,
            808974773: False,
        }
        for dom_id, fifo_status in dct_fifo_stat.items():
            frame = s[s.dom_id == dom_id]
            assert any(get_channel_flags(frame.fifo[0])) == fifo_status

    def test_has_udp_trailer(self):
        s = self.ss.slices[0]
        dct_udp_trailer = {
            806451572: True,
            806455814: True,
            806465101: True,
            806483369: True,
            806487219: True,
            806487226: True,
            806487231: True,
            808432835: True,
            808435278: True,
            808447180: True,
            808447186: True,
        }
        for dom_id, udp_trailer in dct_udp_trailer.items():
            frame = s[s.dom_id == dom_id]
            assert has_udp_trailer(frame.fifo[0]) == udp_trailer

    def test_high_rate_veto(self):
        s = self.ss.slices[0]
        dct_high_rate_veto = {
            808489014: True,
            808489117: False,
            808493910: True,
            808946818: True,
            808951460: True,
            808956908: True,
            808959411: True,
            808961448: True,
            808961480: True,
            808961504: True,
            808961655: False,
            808964815: False,
            808964852: True,
            808969848: False,
            808969857: True,
            808972593: True,
            808972598: True,
            808972698: False,
            808974758: False,
            808974773: True,
            808974811: True,
            808974972: True,
            808976377: True,
            808979567: False,
            808979721: False,
            808979729: False,
            808981510: True,
            808981523: True,
            808981672: False,
            808981812: True,
            808981864: False,
            808982018: False,
        }
        for dom_id, high_rate_veto in dct_high_rate_veto.items():
            frame = s[s.dom_id == dom_id]
            assert any(get_channel_flags(frame.hrv[0])) == high_rate_veto

    def test_max_sequence_number(self):
        s = self.ss.slices[0]
        dct_seq_numbers = {
            808974758: 18,
            808974773: 26,
            808974811: 25,
            808974972: 41,
            808976377: 35,
            808979567: 20,
            808979721: 17,
            808979729: 25,
            808981510: 35,
            808981523: 27,
            808981672: 17,
            808981812: 34,
            808981864: 18,
            808982018: 21,
            808982041: 27,
            808982077: 32,
            808982547: 20,
            808984711: 26,
            808996773: 31,
            808997793: 21,
            809006037: 26,
            809007627: 18,
            809503416: 28,
            809521500: 31,
            809524432: 21,
            809526097: 23,
            809544058: 21,
            809544061: 23,
        }
        for dom_id, max_sequence_number in dct_seq_numbers.items():
            frame = s[s.dom_id == dom_id]
            assert (
                get_udp_max_sequence_number(frame.dq_status[0]) == max_sequence_number
            )

    def test_number_udp_packets(self):
        s = self.ss.slices[0]
        dct_n_packets = {
            808451904: 27,
            808451907: 22,
            808469129: 20,
            808472260: 21,
            808472265: 22,
            808488895: 20,
            808488990: 20,
            808489014: 28,
            808489117: 22,
            808493910: 26,
            808946818: 23,
            808951460: 37,
            808956908: 33,
            808959411: 36,
            808961448: 28,
            808961480: 24,
            808961504: 28,
            808961655: 20,
            808964815: 20,
            808964852: 28,
            808969848: 21,
        }
        for dom_id, n_udp_packets in dct_n_packets.items():
            frame = s[s.dom_id == dom_id]
            assert get_number_udp_packets(frame.dq_status[0]) == n_udp_packets

    def test_hrv_flags(self):
        s = self.ss.slices[0]
        dct_hrv_flags = {
            809524432: [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
            809526097: [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                False,
            ],
            809544058: [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
            809544061: [
                False,
                True,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
        }
        for dom_id, hrv_flags in dct_hrv_flags.items():
            frame = s[s.dom_id == dom_id]
            assert any(
                [a == b for a, b in zip(get_channel_flags(frame.hrv[0]), hrv_flags)]
            )

    def test_fifo_flags(self):
        s = self.ss.slices[0]
        dct_fifo_flags = {
            808982547: [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
            808984711: [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
            808996773: [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
            808997793: [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
            809006037: [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
            808981510: [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                True,
                False,
                False,
                False,
                True,
                False,
                True,
                True,
                True,
                True,
                True,
                True,
                False,
                False,
                True,
                False,
            ],
        }
        for dom_id, fifo_flags in dct_fifo_flags.items():
            frame = s[s.dom_id == dom_id]
            assert any(
                [a == b for a, b in zip(get_channel_flags(frame.fifo[0]), fifo_flags)]
            )

    def test_str(self):
        print(str(self.ss))


class TestGetChannelFlags_Issue59(unittest.TestCase):
    def test_sample_summaryslice_dump(self):
        fieldnames = ["dom_id"]

        for i in range(31):
            fieldnames.append(f"ch{i}")
            fieldnames.append(f"hrvfifo{i}")

        Entry = namedtuple("Entry", fieldnames)

        with open(
            data_path("online/KM3NeT_00000049_00008456.summaryslice-167941.txt")
        ) as fobj:
            ref_entries = [Entry(*list(l.strip().split())) for l in fobj.readlines()]

        r = OnlineReader(
            data_path("online/KM3NeT_00000049_00008456.summaryslice-167941.root")
        )
        summaryslice = r.summaryslices.slices[0]

        for ours, ref in zip(summaryslice, ref_entries):
            assert ours.dom_id == to_num(ref.dom_id)
            fifos = get_channel_flags(ours.fifo)
            hrvs = get_channel_flags(ours.hrv)
            for i in range(31):
                attr = f"ch{i}"
                self.assertAlmostEqual(
                    get_rate(getattr(ours, attr)) / 1000.0,
                    to_num(getattr(ref, attr)),
                    places=1,
                )

                hrvfifo = getattr(ref, f"hrvfifo{i}")
                ref_hrv = bool(int(hrvfifo[0]))
                ref_fifo = bool(int(hrvfifo[1]))
                assert hrvs[i] == ref_hrv
                assert fifos[i] == ref_fifo


class TestGetRate(unittest.TestCase):
    def test_zero(self):
        assert 0 == get_rate(0)

    def test_some_values(self):
        assert 2054 == get_rate(1)
        assert 55987 == get_rate(123)
        assert 1999999 == get_rate(255)

    def test_vectorized_input(self):
        self.assertListEqual([2054], list(get_rate([1])))
        self.assertListEqual([2054, 2111, 2169], list(get_rate([1, 2, 3])))



import binascii
import os
import uproot3
import numpy as np

import numba as nb

TIMESLICE_FRAME_BASKET_CACHE_SIZE = 523 * 1024 ** 2  # [byte]
SUMMARYSLICE_FRAME_BASKET_CACHE_SIZE = 523 * 1024 ** 2  # [byte]
BASKET_CACHE_SIZE = 110 * 1024 ** 2
BASKET_CACHE = uproot3.cache.ThreadSafeArrayCache(BASKET_CACHE_SIZE)

# Parameters for PMT rate conversions, since the rates in summary slices are
# stored as a single byte to save space. The values from 0-255 can be decoded
# using the `get_rate(value)` function, which will yield the actual rate
# in Hz.
MINIMAL_RATE_HZ = 2.0e3
MAXIMAL_RATE_HZ = 2.0e6
RATE_FACTOR = np.log(MAXIMAL_RATE_HZ / MINIMAL_RATE_HZ) / 255

CHANNEL_BITS_TEMPLATE = np.zeros(31, dtype=bool)


@nb.vectorize(
    [nb.int32(nb.int8), nb.int32(nb.int16), nb.int32(nb.int32), nb.int32(nb.int64)]
)
def get_rate(value):  # pragma: no cover
    """Return the rate in Hz from the short int value"""
    if value == 0:
        return 0
    else:
        return MINIMAL_RATE_HZ * np.exp(value * RATE_FACTOR)


@nb.guvectorize(
    "void(i8, b1[:], b1[:])", "(), (n) -> (n)", target="parallel", nopython=True
)
def unpack_bits(value, bits_template, out):  # pragma: no cover
    """Return a boolean array for a value's bit representation.

    This function also accepts arrays as input, the output shape will be
    NxM where N is the number of input values and M the length of the
    ``bits_template`` array, which is just a dummy array, due to the weird
    signature system of numba.

    Parameters
    ----------
    value: int or np.array(int) with shape (N,)
        The binary value of containing the bit information
    bits_template: np.array() with shape (M,)
        The template for the output array, the only important is its shape

    Returns
    -------
    np.array(bool) either with shape (M,) or (N, M)
    """
    for i in range(bits_template.shape[0]):
        out[30 - i] = value & (1 << i) > 0


def get_channel_flags(value):
    """Returns the hrv/fifo flags for the PMT channels (hrv/fifo)

    Parameters
    ----------
    value : int32
        The integer value to be parsed.
    """
    channel_bits = np.bitwise_and(value, 0x7FFFFFFF)
    flags = unpack_bits(channel_bits, CHANNEL_BITS_TEMPLATE)
    return np.flip(flags, axis=-1)


def get_number_udp_packets(value):
    """Returns the number of received UDP packets (dq_status)

    Parameters
    ----------
    value : int32
        The integer value to be parsed.
    """
    return np.bitwise_and(value, 0x7FFF)


def get_udp_max_sequence_number(value):
    """Returns the maximum sequence number of the received UDP packets (dq_status)

    Parameters
    ----------
    value : int32
        The integer value to be parsed.
    """
    return np.right_shift(value, 16)


def has_udp_trailer(value):
    """Returns the UDP Trailer flag (fifo)

    Parameters
    ----------
    value : int32
        The integer value to be parsed.
    """
    return np.any(np.bitwise_and(value, np.left_shift(1, 31)))


class OnlineReader:
    """Reader for online ROOT files"""

    def __init__(self, filename):
        self._fobj = uproot3.open(filename)
        self._filename = filename
        self._events = None
        self._timeslices = None
        self._summaryslices = None
        self._uuid = binascii.hexlify(self._fobj._context.uuid).decode("ascii")

    @property
    def uuid(self):
        return self._uuid

    def close(self):
        self._fobj.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    @property
    def events(self):
        if self._events is None:
            tree = self._fobj["KM3NET_EVENT"]

            headers = tree["KM3NETDAQ::JDAQEventHeader"].array(
                uproot3.interpret(tree["KM3NETDAQ::JDAQEventHeader"], cntvers=True)
            )
            snapshot_hits = tree["snapshotHits"].array(
                uproot3.asjagged(
                    uproot3.astable(
                        uproot3.asdtype(
                            [
                                ("dom_id", ">i4"),
                                ("channel_id", "u1"),
                                ("time", "<u4"),
                                ("tot", "u1"),
                            ]
                        )
                    ),
                    skipbytes=10,
                )
            )
            triggered_hits = tree["triggeredHits"].array(
                uproot3.asjagged(
                    uproot3.astable(
                        uproot3.asdtype(
                            [
                                ("dom_id", ">i4"),
                                ("channel_id", "u1"),
                                ("time", "<u4"),
                                ("tot", "u1"),
                                (" cnt", "u4"),
                                (" vers", "u2"),
                                ("trigger_mask", ">u8"),
                            ]
                        )
                    ),
                    skipbytes=10,
                )
            )
            self._events = OnlineEvents(headers, snapshot_hits, triggered_hits)
        return self._events

    @property
    def timeslices(self):
        if self._timeslices is None:
            self._timeslices = Timeslices(self._fobj)
        return self._timeslices

    @property
    def summaryslices(self):
        if self._summaryslices is None:
            self._summaryslices = SummarySlices(self._fobj)
        return self._summaryslices


class SummarySlices:
    """A wrapper for summary slices"""

    def __init__(self, fobj):
        self._fobj = fobj
        self._slices = None
        self._headers = None
        self._rates = None
        self._ch_selector = ["ch{}".format(c) for c in range(31)]

    @property
    def headers(self):
        if self._headers is None:
            self._headers = self._read_headers()
        return self._headers

    @property
    def slices(self):
        if self._slices is None:
            self._slices = self._read_summaryslices()
        return self._slices

    @property
    def rates(self):
        if self._rates is None:
            self._rates = self.slices[["dom_id"] + self._ch_selector]
        return self._rates

    def _read_summaryslices(self):
        """Reads a lazyarray of summary slices"""
        tree = self._fobj[b"KM3NET_SUMMARYSLICE"][b"KM3NET_SUMMARYSLICE"]
        return tree[b"vector<KM3NETDAQ::JDAQSummaryFrame>"].lazyarray(
            uproot3.asjagged(
                uproot3.astable(
                    uproot3.asdtype(
                        [
                            ("dom_id", "i4"),
                            ("dq_status", "u4"),
                            ("hrv", "u4"),
                            ("fifo", "u4"),
                            ("status3", "u4"),
                            ("status4", "u4"),
                        ]
                        + [(c, "u1") for c in self._ch_selector]
                    )
                ),
                skipbytes=10,
            ),
            basketcache=uproot3.cache.ThreadSafeArrayCache(
                SUMMARYSLICE_FRAME_BASKET_CACHE_SIZE
            ),
        )

    def _read_headers(self):
        """Reads a lazyarray of summary slice headers"""
        tree = self._fobj[b"KM3NET_SUMMARYSLICE"][b"KM3NET_SUMMARYSLICE"]
        return tree[b"KM3NETDAQ::JDAQSummarysliceHeader"].lazyarray(
            uproot3.interpret(tree[b"KM3NETDAQ::JDAQSummarysliceHeader"], cntvers=True)
        )

    def __str__(self):
        return "Number of summaryslices: {}".format(len(self.headers))


class Timeslices:
    """A simple wrapper for timeslices"""

    def __init__(self, fobj):
        self._fobj = fobj
        self._timeslices = {}
        self._read_streams()

    def _read_streams(self):
        """Read the L0, L1, L2 and SN streams if available"""
        streams = set(
            s.split(b"KM3NET_TIMESLICE_")[1].split(b";")[0]
            for s in self._fobj.keys()
            if b"KM3NET_TIMESLICE_" in s
        )
        for stream in streams:
            tree = self._fobj[b"KM3NET_TIMESLICE_" + stream][
                b"KM3NETDAQ::JDAQTimeslice"
            ]
            headers = tree[b"KM3NETDAQ::JDAQTimesliceHeader"][b"KM3NETDAQ::JDAQHeader"][
                b"KM3NETDAQ::JDAQChronometer"
            ]
            if len(headers) == 0:
                continue
            superframes = tree[b"vector<KM3NETDAQ::JDAQSuperFrame>"]
            hits_dtype = np.dtype([("pmt", "u1"), ("tdc", "<u4"), ("tot", "u1")])
            hits_buffer = superframes[
                b"vector<KM3NETDAQ::JDAQSuperFrame>.buffer"
            ].lazyarray(
                uproot3.asjagged(
                    uproot3.astable(uproot3.asdtype(hits_dtype)), skipbytes=6
                ),
                basketcache=uproot3.cache.ThreadSafeArrayCache(
                    TIMESLICE_FRAME_BASKET_CACHE_SIZE
                ),
            )
            self._timeslices[stream.decode("ascii")] = (
                headers,
                superframes,
                hits_buffer,
            )
            setattr(
                self,
                stream.decode("ascii"),
                TimesliceStream(headers, superframes, hits_buffer),
            )

    def stream(self, stream, idx):
        ts = self._timeslices[stream]
        return Timeslice(*ts, idx, stream)

    def __str__(self):
        return "Available timeslice streams: {}".format(
            ", ".join(s for s in self._timeslices.keys())
        )

    def __repr__(self):
        return str(self)


class TimesliceStream:
    def __init__(self, headers, superframes, hits_buffer):
        # self.headers = headers.lazyarray(
        #     uproot3.asjagged(uproot3.astable(
        #         uproot3.asdtype(
        #             np.dtype([('a', 'i4'), ('b', 'i4'), ('c', 'i4'),
        #                       ('d', 'i4'), ('e', 'i4')]))),
        #                     skipbytes=6),
        #     basketcache=uproot3.cache.ThreadSafeArrayCache(
        #         TIMESLICE_FRAME_BASKET_CACHE_SIZE))
        self.headers = headers
        self.superframes = superframes
        self._hits_buffer = hits_buffer

    # def frames(self):
    #     n_hits = self._superframe[
    #         b'vector<KM3NETDAQ::JDAQSuperFrame>.numberOfHits'].lazyarray(
    #             basketcache=BASKET_CACHE)[self._idx]
    #     module_ids = self._superframe[
    #         b'vector<KM3NETDAQ::JDAQSuperFrame>.id'].lazyarray(basketcache=BASKET_CACHE)[self._idx]
    #     idx = 0
    #     for module_id, n_hits in zip(module_ids, n_hits):
    #         self._frames[module_id] = hits_buffer[idx:idx + n_hits]
    #         idx += n_hits


class Timeslice:
    """A wrapper for a timeslice"""

    def __init__(self, header, superframe, hits_buffer, idx, stream):
        self.header = header
        self._frames = {}
        self._superframe = superframe
        self._hits_buffer = hits_buffer
        self._idx = idx
        self._stream = stream
        self._n_frames = None

    @property
    def frames(self):
        if not self._frames:
            self._read_frames()
        return self._frames

    def _read_frames(self):
        """Populate a dictionary of frames with the module ID as key"""
        hits_buffer = self._hits_buffer[self._idx]
        n_hits = self._superframe[
            b"vector<KM3NETDAQ::JDAQSuperFrame>.numberOfHits"
        ].lazyarray(basketcache=BASKET_CACHE)[self._idx]
        try:
            module_ids = self._superframe[
                b"vector<KM3NETDAQ::JDAQSuperFrame>.id"
            ].lazyarray(basketcache=BASKET_CACHE)[self._idx]
        except KeyError:
            module_ids = (
                self._superframe[
                    b"vector<KM3NETDAQ::JDAQSuperFrame>.KM3NETDAQ::JDAQModuleIdentifier"
                ]
                .lazyarray(
                    uproot3.asjagged(
                        uproot3.astable(uproot3.asdtype([("dom_id", ">i4")]))
                    ),
                    basketcache=BASKET_CACHE,
                )[self._idx]
                .dom_id
            )

        idx = 0
        for module_id, n_hits in zip(module_ids, n_hits):
            self._frames[module_id] = hits_buffer[idx : idx + n_hits]
            idx += n_hits

    def __len__(self):
        if self._n_frames is None:
            self._n_frames = len(
                self._superframe[b"vector<KM3NETDAQ::JDAQSuperFrame>.id"].lazyarray(
                    basketcache=BASKET_CACHE
                )[self._idx]
            )
        return self._n_frames

    def __str__(self):
        return "{} timeslice with {} frames.".format(self._stream, len(self))

    def __repr__(self):
        return "<{}: {} entries>".format(self.__class__.__name__, len(self.header))


class OnlineEvents:
    """A simple wrapper for online events"""

    def __init__(self, headers, snapshot_hits, triggered_hits):
        self.headers = headers
        self.snapshot_hits = snapshot_hits
        self.triggered_hits = triggered_hits

    def __getitem__(self, item):
        return OnlineEvent(
            self.headers[item], self.snapshot_hits[item], self.triggered_hits[item]
        )

    def __len__(self):
        return len(self.headers)

    def __str__(self):
        return "Number of events: {}".format(len(self.headers))

    def __repr__(self):
        return str(self)


class OnlineEvent:
    """A wrapper for a online event"""

    def __init__(self, header, snapshot_hits, triggered_hits):
        self.header = header
        self.snapshot_hits = snapshot_hits
        self.triggered_hits = triggered_hits

    def __str__(self):
        return "Online event with {} snapshot hits and {} triggered hits".format(
            len(self.snapshot_hits), len(self.triggered_hits)
        )

    def __repr__(self):
        return str(self)
